/*
 * icsp.c - part of USBasp
 *
 * Autor..........: Yoko Ono
 * Description....: Provides functions for communication/programming
 *                  over ISP interface
 * Licence........: GNU GPL v2 (see Readme.txt)
 * Creation Date..: 2005-02-23
 * Last change....: 2010-01-19
 */

#include <avr/io.h>
#include "isp.h"
#include "clock.h"
#include "usbasp.h"

#include <util/delay.h>

#include "icsp.h"


#define spiHWdisable() SPCR = 0

// note: all data is sampled on the falling edge
//
// AVR ISP -> PIC ICSP pin identifications:
//
//  nMCLR	- RESET
//  PGC		- SCK
//  PGD		- MOSI

//#define clk_wait() clockWait(1)
#define clk_wait(x) _delay_ms(1)
#define clockWait(x) _delay_ms(1)

//#define MCLR_INV
#define CLR_PGD

/*
#ifdef MCLR_INV
#define MCLR_HIGH() ISP_OUT &= ~(1 << ISP_RST); uart_putc('M')
#define MCLR_LOW()  ISP_OUT |=  (1 << ISP_RST); uart_putc('m')
#else
#define MCLR_HIGH() ISP_OUT |=  (1 << ISP_RST); uart_putc('M')
#define MCLR_LOW()  ISP_OUT &= ~(1 << ISP_RST); uart_putc('m')
#endif
*/

#ifdef MCLR_INV
#define MCLR_HIGH() ISP_OUT &= ~(1 << ISP_RST);
#define MCLR_LOW()  ISP_OUT |=  (1 << ISP_RST);
#else
#define MCLR_HIGH() ISP_OUT |=  (1 << ISP_RST);
#define MCLR_LOW()  ISP_OUT &= ~(1 << ISP_RST);
#endif

/*
#define SEND_BIT(bit) { \
	if ((bit)) { ISP_OUT |=  (1 << ISP_MOSI); uart_putc('1'); } \
	else       { ISP_OUT &= ~(1 << ISP_MOSI); uart_putc('0'); } \
	_delay_ms(1); \
	ISP_OUT |= (1 << ISP_SCK); \
	_delay_ms(1); \
	ISP_OUT &= ~(1 << ISP_SCK); \
	_delay_ms(1); }
*/
#define SEND_BIT(bit) { \
	if ((bit)) ISP_OUT |=  (1 << ISP_MOSI); \
	else       ISP_OUT &= ~(1 << ISP_MOSI); \
	ISP_OUT |= (1 << ISP_SCK); \
	ISP_OUT &= ~(1 << ISP_SCK); }

//#define NL() uart_putc('\n');
#define NL()

void icsp_init(uint32_t magic)
{	
	// disable hardware SPI
//	spiHWdisable();
	
	// all ISP pins are inputs before
	// now set output pins
	ISP_DDR |= (1 << ISP_RST) | (1 << ISP_SCK) | (1 << ISP_MOSI)  | (1 << ISP_MISO);

	MCLR_LOW();						// nMCLR low
	ISP_OUT &= ~(1 << ISP_SCK);		//   PGC low
	ISP_OUT &= ~(1 << ISP_MOSI);	//   PGD low
	_delay_ms(10);
	MCLR_HIGH();					// nMCLR high
	_delay_ms(10);

	MCLR_LOW();						// nMCLR low
	_delay_ms(10);					// nMCLR P19 low

//	icsp_sendbits(0x4D434850, 32); // 0100 1101 0100 0011 ...
	icsp_sendbits(0x4d, 8);
	icsp_sendbits(0x43, 8);
	icsp_sendbits(0x48, 8);
	icsp_sendbits(0x50, 8);
	NL();
	
	_delay_ms(10);					// P20

	MCLR_HIGH();					// nMCLR high
	_delay_ms(10);					// P12
}

void icsp_done(void)
{
	ISP_OUT &= ~(1 << ISP_SCK);		//   PGC low
	ISP_OUT &= ~(1 << ISP_MOSI);	//   PGD low
	
	// do a reset
	MCLR_LOW();
	_delay_ms(10);
	MCLR_HIGH();
}

// synonymous for sendbits32
// (MSB first, all other functions do LSB first)
void icsp_sendbits(uint32_t bits, uint8_t nbits)
{
	uchar n;

	// raise clock ... set data bit ... lower clock
	for (n=0; n<nbits; n++) SEND_BIT((bits >> (nbits-1-n)) & 1);

	ISP_OUT &= ~(1 << ISP_MOSI);		// PGD low
}

void icsp_send16(uint8_t cmd, uint16_t data)
{
	uchar n;

	// 4 bit command
	for (n=0; n<4; n++) SEND_BIT(cmd & (1 << n));

#ifdef CLR_PGD
	ISP_OUT &= ~(1 << ISP_MOSI);		// PGD low
#endif

	// 16 bit data
	for (n=0; n<16; n++) SEND_BIT(data & (1 << n));

#ifdef CLR_PGD
	ISP_OUT &= ~(1 << ISP_MOSI);
#endif
}

uint8_t icsp_read8(uint8_t cmd)
{
	uint8_t n, ret;

	// 4 bit command
	for (n=0; n<4; n++) SEND_BIT(cmd & (1 << n));
	ISP_OUT &= ~(1 << ISP_MOSI);

	// 8 clocks
	for (n=0; n<8; n++) SEND_BIT(0);
	
	ISP_DDR &= ~(1 << ISP_MOSI);	// make MOSI(PGD) input
	ret = 0;

	// raise clock ... lower clock .. read bit
	for (n=0; n<8; n++) {
		ISP_OUT |= (1 << ISP_SCK);
		ISP_OUT &= ~(1 << ISP_SCK);
		if (ISP_IN & (1 << ISP_MOSI)) ret |= (1 << n);
	}
	
	ISP_DDR |= (1 << ISP_MOSI);	// make MOSI(PGD) output again

	return ret;
}

void icsp_special_nop(uint8_t hi)
{
	uint8_t n;
	
	// 3 clocks
	for (n=0; n<3; n++) SEND_BIT(0);
	
	// 4th clock is special
	ISP_OUT |= (1 << ISP_SCK);
	if (!hi) {
		ISP_OUT &= ~(1 << ISP_SCK);
		_delay_ms(700);	// P11
	}
	else {
		_delay_ms(60);	// max(P9, P10)
		ISP_OUT &= ~(1 << ISP_SCK);
		_delay_ms(1);
	}
	
	// 16bit payload
	for (n=0; n<16; n++) SEND_BIT(0);

#ifdef CLR_PGD
	ISP_OUT &= ~(1 << ISP_MOSI);
#endif
}

#ifdef __PIC32_2WIRE_JTAG__

// two phase 2 wire jtag
#define SEND_ONE_2P(tms, tdi) { \
	if ((tdi)) ISP_OUT |=  (1 << ISP_MOSI); \
	else       ISP_OUT &= ~(1 << ISP_MOSI); \
	ISP_OUT |= (1 << ISP_SCK); \
	ISP_OUT &= ~(1 << ISP_SCK); \
	if ((tms)) ISP_OUT |=  (1 << ISP_MOSI); \
	else       ISP_OUT &= ~(1 << ISP_MOSI); \
	ISP_OUT |= (1 << ISP_SCK); \
	ISP_OUT &= ~(1 << ISP_SCK); }

// input data is sampled on the falling edge of PGC
// output data (TDO) is driven on the falling edge of PGC

// goto Run-Test/Idle state
void SetMode(uint8_t nbits, uint32_t bits)
{
	
}

#endif // __PIC32_2WIRE_JTAG__
